#include <errno.h>

#include "task.h"
#include "bitmap.h"

/* Should be declared in <string.h> and <assert.h> */
#define memcpy __builtin_memcpy
#define assert(x)

/* For simplicity, use fix size, reallocation could be added later.
 * Should be a power of two.*/
#define TASK_MAX 1024
#define THREAD_MAX 1024
#define HANDLE_MAX 128

/* Our subsystem */
static l4_word_t task_subsystem;
/* Our task id, lower task id:s are reserved */
static l4_word_t task_reserved;

/* kip locations for all tasks we create */
static l4_fpage_t kip_fpage;

static int ever_registered;

/* Alternative task object implementation */
struct task_object
{  
  enum task_state state;
  
  /* An inactive thread (tid zero) used to identify the space. This
     also gives us the global taskid. */
  l4_thread_id_t space;

  /* Only a single accounting id */
  l4_word_t accounting_id;
  
  /* Number of active threads */
  l4_word_t nthreads;

  /* Number of threads that we have utc blocks for */
  l4_word_t thread_max;
  l4_word_t thread_bitmap[THREAD_MAX / L4_BITS_PER_WORD];

  struct task_handle handles[HANDLE_MAX];

  /* Number of tasks and handles that refer to us. */
  l4_word_t nreferrers;
  /* Set if sending a death notification to this task fails */
  int out_of_sync;

  /* Notify thread */
  l4_thread_id_t notify;
};

/* Indexed by task id, so some entries are unused.

   FIXME: We waste quite a lot of memory here. When we have dynamic
   allocation, this array is still needed to get the taskid to task
   struct mapping, but it can be replaced by a struct task_object *[],
   i.e. an array of pointers to independently allocated task_object.
   The poitner array itself can also be allocated dynamically, to save
   space for systems with few tasks running. */

static struct task_object
task_list[TASK_MAX];

void
task_init(void)
{
  /* Always returns a global id */
  l4_thread_id_t self = l4_myself();

  task_subsystem = l4_thread_no(self) >> SUBSYSTEM_SHIFT;
  task_reserved = l4_version(self);
  
  /* Redundant if linker/loader does its job */
  memset(task_list, 0, sizeof(task_list));
  
  kip_fpage = l4_fpage_log2 (KIP_LOCATION,
			     l4_kip_area_size_log2 ());

  ever_registered = 0;
}

/* Returns NULL on failure */
static struct task_object *
task_alloc (void)
{
  static unsigned count = 0;
  unsigned id;

  for (id = count + 1; id != count; id = (id + 1) % TASK_MAX)
    {
      if ( (id <= task_reserved) || (id % 64 == 0))
	continue;
      if (task_list[id].state == TASK_FREE)
	{
	  task_list[id].state = TASK_EMPTY;
	  task_list[id].space
	    = l4_global_id(task_subsystem << SUBSYSTEM_SHIFT, id);
	  count = id;
	  return task_list + id;
	}
    }
  return NULL;
}

static void
task_free(struct task_object *task)
{
    task->state = TASK_FREE;  
}

static void
task_maybe_free(struct task_object *task)
{
  if (task->state == TASK_ZOMBIE
      && task->nreferrers == 0)
    task_free (task);
}

void
task_incref(struct task_object *task)
{
  assert(task->state != TASK_FREE);
  task->nreferrers++;
}

void
task_decref(struct task_object *task)
{
  assert(task->state != TASK_FREE);
  assert(task->nreferrers);
  task->nreferrers--;

  task_maybe_free(task);
}

/* Returns task id. */
l4_word_t
taskid_by_thread(l4_thread_id_t thread)
{
  l4_word_t id;
  
  if (!l4_is_global_id(thread)
      || (l4_thread_no(thread) >> SUBSYSTEM_SHIFT) != task_subsystem)
    return 0;

  id = l4_version(thread);
  if (task_reserved < id && id < TASK_MAX)
    return id;

  return 0;
}

struct task_object *
task_by_thread(l4_thread_id_t thread)
{
  l4_word_t id = taskid_by_thread (thread);
  if (!id)
    return NULL;
  
  assert (task_list[id].state == TASK_ALIVE);
  return task_list + id;  
}

const struct task_handle *
task_handle_lookup(struct task_object *owner, l4_word_t id)
{
  if (id < HANDLE_MAX
      && (owner->handles[id].type == HANDLE_REF
	  || owner->handles[id].type == HANDLE_CONTROL))
    {
      assert (owner->handles[id].task->state != TASK_FREE);
      return owner->handles + id;
    }
  return NULL;
}

const struct task_object *
task_by_handle(struct task_object *owner, l4_word_t id)
{
  const struct task_handle *handle = task_handle_lookup(owner, id);
  if (handle
      && (handle->type == HANDLE_REF
	  || handle->type == HANDLE_CONTROL))
    return handle->task;
  else
    return NULL;
}

struct task_object *
task_by_control_handle(struct task_object *owner, l4_word_t id)
{
  const struct task_handle *handle = task_handle_lookup(owner, id);
  if (handle
      && handle->type == HANDLE_CONTROL)
    return handle->task;
  else
    return NULL;
}

int
task_manager_p(struct task_object *owner, l4_word_t id)
{
  return (id < HANDLE_MAX
	  && owner->handles[id].type == HANDLE_MANAGER);
}

/* Thread allocation */

static inline int
task_thread_used_p(const struct task_object *task, l4_word_t tid)
{
  return 0 < tid && tid < THREAD_MAX && bit_p(tid, task->thread_bitmap);
}

/* Does not mark the thread as allocated */
static l4_word_t
task_thread_alloc(struct task_object *task, void **utc)
{
  /* Again, an unefficient linear search */
  unsigned tid;

  /* tid zero is reserved for the inactive thread used as the space
   * specifier */
  for (tid = 1; tid < task->thread_max; tid++)
    if (!task_thread_used_p(task, tid))
      {
	/* A free thread number */
	*utc = (void *) (UTC_LOCATION + tid * l4_utcb_size());
	return tid;
      }
  return 0;
}

static void
task_thread_use(struct task_object *task, l4_word_t tid)
{
  bit_set(tid, task->thread_bitmap);
  task->nthreads++;
}

static void
task_thread_unuse(struct task_object *task, l4_word_t tid)
{
  assert(task->nthreads);
  bit_clear(tid, task->thread_bitmap);
  task->nthreads--;
}

static inline l4_thread_id_t
make_thread_id(struct task_object *task, l4_word_t tid)
{
  return l4_global_id(tid + (task_subsystem << SUBSYSTEM_SHIFT),
		      l4_version(task->space)); 
}

/* Checks that THREAD really is a thread in the given task,
   and returns the tid. */
static inline l4_word_t
check_thread_id(const struct task_object *task, l4_thread_id_t thread)
{
  l4_word_t tid;
  
  if (l4_version(task->space) != taskid_by_thread(thread))
    return 0;
  
  tid = l4_thread_no(thread) % (1UL << SUBSYSTEM_SHIFT);

  if (task_thread_used_p (task, tid))
    return tid;
  else
    return 0;
}


/* Handle allocation */
static struct task_handle *
task_handle_alloc (struct task_object *task, enum task_handle_type type,
		   l4_word_t *id)
{
  l4_word_t i;
  
  for (i = 0; i < HANDLE_MAX; i++)
    {
      if (task->handles[i].type == HANDLE_FREE)
	{
	  struct task_handle *handle = task->handles + i;
	  *id = i;
	  
	  handle->type = type;
	  handle->refs = 1;
	  handle->task = NULL;

	  return handle;
	}
    }
  return NULL;
}

static void
task_handle_free (struct task_handle *handle)
{
  handle->type = HANDLE_FREE;
}

l4_word_t
task_add_cap (struct task_object *task,
	      enum task_handle_type type,
	      struct task_object *subject,
	      l4_word_t *id)
{
  struct task_handle *handle;
  
  int i;

  for (i = 0; i < HANDLE_MAX; i++)
    {
      if (task->handles[i].type == type
	  && task->handles[i].task == subject)
	{
	  *id = i;
	  task->handles[i].refs++;
	  return 0;
	}
    }

  handle = task_handle_alloc (task, type, id);
  if (!handle)
    return ENOMEM;

  handle->task = subject;
  return 0;
}

/* Syscall wrappers */
#define WORTEL_MSG_SPACE_CONTROL 10
#define WORTEL_MSG_THREAD_CONTROL 11

static l4_word_t
wortel_space_control (l4_thread_id_t space, l4_word_t control,
		      l4_fpage_t kip_area, l4_fpage_t utcb_area,
		      l4_thread_id_t redirector, l4_word_t *old_control)
{
  l4_msg_t msg;
  l4_msg_tag_t tag;
  
  l4_msg_clear (&msg);
  l4_set_msg_label (&msg, WORTEL_MSG_SPACE_CONTROL);

  l4_msg_append_word (&msg, TASK_WORTEL_CAP);

  l4_msg_append_word (&msg, space.raw);
  l4_msg_append_word (&msg, control);
  l4_msg_append_word (&msg, kip_area.raw);
  l4_msg_append_word (&msg, utcb_area.raw);
  l4_msg_append_word (&msg, redirector.raw);

  l4_msg_load (&msg);
  /* FIXME: Hard coded thread ID.  */
  tag = l4_call (l4_global_id (l4_thread_user_base () + 2, 1));
  if (l4_ipc_failed (tag))
    return 0;

  if (l4_label (tag) == 0)
    return 0;

  if (l4_untyped_words (tag) != 1
      || l4_typed_words (tag) != 0)
    panic ("Invalid format of space_control reply");
  
  l4_msg_store (tag, &msg);

  *old_control = l4_msg_word (&msg, 0);

  return 1;  
}

static l4_word_t
wortel_thread_control (l4_thread_id_t dest, l4_thread_id_t space,
		       l4_thread_id_t scheduler, l4_thread_id_t pager,
		       void *utcb_loc)
{
  l4_msg_t msg;
  l4_msg_tag_t tag;
  
  l4_msg_clear (&msg);
  l4_set_msg_label (&msg, WORTEL_MSG_THREAD_CONTROL);

  l4_msg_append_word (&msg, TASK_WORTEL_CAP);

  l4_msg_append_word (&msg, dest.raw);
  l4_msg_append_word (&msg, space.raw);
  l4_msg_append_word (&msg, scheduler.raw);
  l4_msg_append_word (&msg, pager.raw);
  l4_msg_append_word (&msg, (l4_word_t) utcb_loc);

  l4_msg_load (&msg);
  /* FIXME: Hard coded thread ID.  */
  tag = l4_call (l4_global_id (l4_thread_user_base () + 2, 1));
  if (l4_ipc_failed (tag))
    return 0;

  return l4_label (tag);
}


/* Notifications */

static void
task_zombify(struct task_object *task)
{
  l4_word_t i;
  l4_word_t o;
  
  assert(task->state == TASK_ALIVE);
  assert(!task->nthreads);

  /* Delete the address space. Ignores any error. */
  (void) wortel_thread_control(task->space, l4_nilthread, l4_nilthread,
			       l4_nilthread, (void *) -1);
  
  /* First clean up any transaction references the dead task had. The
     call may deallocate some zombie tasks, so we must keep our state
     as alive during this call. */
  transaction_note_death (task);
  task->state = TASK_ZOMBIE;

  /* References we have to other tasks should be deleted */
  for (i = 0; i < HANDLE_MAX; i++)
    {
      if (task->handles[i].type == HANDLE_REF
	  || task->handles[i].type == HANDLE_CONTROL)
	{
	  task->handles[i].task->nreferrers--;
	  task_maybe_free (task->handles[i].task);
	}
    }

  /* Send notifications to tasks that refer to us */
  for (o = task_reserved + 1; o < TASK_MAX; o++)
    {
      if (task_list[o].state == TASK_ALIVE
	  && !l4_is_nil_thread(task_list[o].notify))
	{
	  for (i = 0; i < HANDLE_MAX; i++)
	    if ((task_list[o].handles[i].type == HANDLE_REF
		 || task_list[o].handles[i].type == HANDLE_CONTROL)
		&& task_list[o].handles[i].task == task)
	      {
		l4_msg_t msg;
	      
		/* We send a single notification, no matter if the task
		   had a reference or a control handle or both. We don't
		   do any retransmissions, but we record failures. A
		   task could find out that it is out of sync and
		   request a list of all handles it has to zombie
		   tasks. */
		
		l4_msg_clear (&msg);
		l4_set_msg_label (&msg, TASK_DEATH_NOTIFICATION);
		l4_msg_append_word (&msg, l4_version(task->space));
		l4_msg_load (&msg);

		if (l4_ipc_failed(l4_reply(task_list[o].notify)))
		  task_list[o].out_of_sync = 1;
	      }
	}
    }
  
  task_maybe_free(task);
}

/* Operations */
l4_word_t
task_create(struct task_object *creator, l4_word_t thread_max,
	    l4_word_t *control_id, l4_word_t *ref_id)
{
  struct task_object *task;
  struct task_handle *control;
  struct task_handle *ref;
  
  l4_word_t old;

  task = task_alloc ();
  if (!task)
    return ENOMEM;

  control = task_handle_alloc (creator, HANDLE_CONTROL, control_id);
  if (!control)
    {
      task_free (task);
      return ENOMEM;
    }

  ref = task_handle_alloc (task, HANDLE_REF, ref_id);
  if (!ref)
    {
      task_free (task);
      task_handle_free (control);
      return ENOMEM;
    }
  
  if (!thread_max || thread_max > THREAD_MAX)
    thread_max = THREAD_MAX;
  
  /* Create an address space */
  if (!wortel_thread_control(/* thread, space, scheduler */
	                     task->space, task->space, task->space,
			     /* pager, utc */
			     l4_nilthread, (void *) -1))
    {
      task_free (task);
      task_handle_free (control);
      task_handle_free (ref);
      return ENOMEM;
    }
  
  if (!wortel_space_control(task->space, 0,
			    kip_fpage,
			    l4_fpage(UTC_LOCATION,
				     /* FIXME: Alignment? */
				     thread_max * l4_utcb_size()),
			    l4_anythread, &old))
    {
      /* Ignores any error */
      wortel_thread_control(/* thread, space, scheduler */
	                    task->space, l4_nilthread, task->space,
			    /* pager, utc */
			    l4_nilthread, (void *) -1);
      task_free (task);
      task_handle_free (control);
      task_handle_free (ref);
      return ENOMEM;
    }

  task->thread_max = thread_max;

  task->accounting_id = creator->accounting_id;

  control->task = task;
  ref->task = creator;

  task_incref (task);
  task_incref (creator);
  
  return 0;
}

l4_word_t
task_destroy(struct task_object *task)
{
  l4_word_t tid;

  for (tid = 1; tid < THREAD_MAX; tid++)
    {
      if (task_thread_used_p(task, tid))
	{
	  l4_thread_id_t thread = make_thread_id(task, tid);
	  
	  if (!wortel_thread_control(thread, l4_nilthread, l4_nilthread,
				     l4_nilthread, (void *) -1))
	    panic("task_destroy: wortel_thread_control failed.\n");

	  task_thread_unuse(task, tid);
	}
    }
  task_zombify(task);
  return 0;
}

l4_word_t
task_release(struct task_object *task, l4_word_t id)
{
  struct task_handle *handle;
  
  if (id >= HANDLE_MAX)
    return EINVAL;

  handle = task->handles + id;
  switch (handle->type)
    {
    case HANDLE_REF:
    case HANDLE_CONTROL:
      task_decref (handle->task);
      /* Fall through */
    case HANDLE_MANAGER:
      handle->type = HANDLE_FREE;
      return 0;
    default:
      return EINVAL;
    }
}

l4_word_t
task_query (const struct task_object *task,
	    l4_word_t *taskid, l4_word_t *state)
{
  *taskid = l4_version (task->space);
  *state = task->state;
  return 0;
}

l4_word_t
task_group_destroy(l4_word_t accounting_id)
{
  l4_word_t id;

  for (id = task_reserved + 1; id < TASK_MAX; id++)
    if (task_list[id].state == TASK_ALIVE
	&& task_list[id].accounting_id == accounting_id)
      {
	l4_word_t err = task_destroy(task_list + id);
	if (err)
	  /* Can this happen? */
	  return err;
      }

  return 0;
}

l4_word_t
task_thread_create(struct task_object *task, l4_thread_id_t pager,
		   l4_thread_id_t *new_thread)
{
  l4_word_t tid;
  l4_thread_id_t thread;
  void *utc;

  if (task->state == TASK_ZOMBIE)
    return EPERM;
  
  tid = task_thread_alloc(task, &utc);
  if (!tid)
    return ENOMEM;

  thread = make_thread_id(task, tid);
  
  if (!wortel_thread_control(/* thread, space, scheduler */
	                     thread, task->space, thread,
			     /* pager, utc */
			     pager, utc))
    return ENOMEM;

  task_thread_use(task, tid);
  task->state = TASK_ALIVE;
  
  *new_thread = thread;
  return 0;
}

l4_word_t
task_thread_destroy(struct task_object *task, l4_thread_id_t victim)
{
  l4_word_t tid = check_thread_id(task, victim);

  if (!tid)
    return ESRCH;
  
  if (!wortel_thread_control(victim, l4_nilthread, l4_nilthread,
			     l4_nilthread, (void *) -1))
    return EPERM;

  task_thread_unuse(task, tid);
  if (!task->nthreads)
    task_zombify(task);
  
  return 0;
}


/* Task manager */

l4_word_t
task_register_manager (struct task_object *task, l4_word_t *id)
{
  struct task_handle *handle;
  
  if (ever_registered)
    return EPERM;

  handle = task_handle_alloc (task, HANDLE_MANAGER, id);
  if (!handle)
    return ENOMEM;

  ever_registered = 1;
  return 0;
}
