
#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>

#include "task.h"

/* The program name.  */
char program_name[] = "task";


/* A request with n arguments */
#define CASE(request, n)			\
case request:					\
  if (tag.typed != 0 || tag.untyped != (n)+1)	\
    { err = EINVAL; goto fail; }

/* Main function for each server thread. */
static void
serve(void)
{
  for (;;)
    {
      l4_word_t err;
      l4_msg_tag_t tag;
      l4_msg_t msg;
      l4_word_t label;
      l4_thread_id_t caller;
      struct task_object *caller_task;

      debug("task-server server loop.\n");
      
      tag = l4_wait(&caller);
      printf("task-server received tag %x\n", tag.raw);
      
      if (l4_ipc_failed(tag))
	continue;

      label = l4_label(tag);
      
      /* Checks that the caller is in the right subsystem, and has a
	 sane task id */
      if (!(caller_task = task_by_thread(caller)))
	/* Just drop the request on the floor. */
	continue;

      switch(label)
	{
	  CASE (TASK_CREATE, 1)
	    {
	      l4_word_t thread_max;
	      
	      /* Control handle for new task */
	      l4_word_t control;
	      /* Reference to the old task */
	      l4_word_t ref;
	      
	      l4_msg_store (tag, &msg);
	      thread_max = l4_msg_word(&msg, 1);

	      err = task_create(caller_task, thread_max, &control, &ref);

	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      l4_msg_append_word (&msg, control);
	      l4_msg_append_word (&msg, ref);

	      break;
	    }

	  CASE (TASK_DESTROY, 0)
	    {
	      l4_word_t handle;
	      struct task_object *target;

	      l4_msg_store (tag, &msg);
	      handle = l4_msg_word(&msg, 1);

	      target = task_by_control_handle(caller_task, handle);
	      if (!target)
		{ err = EPERM; goto fail; }

	      err = task_destroy(target);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      break;
	    }

	  CASE (TASK_RELEASE, 0)
	    {
	      l4_word_t handle;

	      l4_msg_store (tag, &msg);
	      handle = l4_msg_word (&msg, 1);

	      err = task_release(caller_task, handle);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      break;
	      
	    }

	  CASE (TASK_QUERY, 0)
	    {
	      const struct task_object *target;
	      l4_word_t handle;
	      l4_word_t id;
	      l4_word_t state;

	      l4_msg_store (tag, &msg);
	      handle = l4_msg_word (&msg, 1);

	      target = task_by_handle(caller_task, handle);
	      if (!target)
		{ err = EPERM; goto fail; }

	      err = task_query(target, &id, &state);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      l4_msg_append_word (&msg, id);
	      l4_msg_append_word (&msg, state);
	      break;
	    }
	  
	  CASE (TASK_GROUP_DESTROY, 2)
	    {
	      l4_word_t handle;
	      l4_word_t accounting_id;

	      l4_msg_store (tag, &msg);
	      handle = l4_msg_word (&msg, 1);
	      accounting_id = l4_msg_word (&msg, 2);

	      if (!task_manager_p (caller_task, handle))
		{ err = EPERM; goto fail; }
	      
	      err = task_group_destroy(accounting_id);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      break;
	    }
	  
	  CASE (TASK_THREAD_CREATE, 2)
	    {
	      struct task_object *target;
	      l4_word_t handle;
	      l4_thread_id_t pager;
	      
	      l4_thread_id_t new_thread;
	      
	      l4_msg_store (tag, &msg);
	      handle = l4_msg_word (&msg, 1);
	      pager.raw = l4_msg_word (&msg, 2);

	      target = task_by_control_handle(caller_task, handle);
	      if (!target)
		{ err = EPERM; goto fail; }

	      err = task_thread_create(target, pager, &new_thread);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      l4_msg_append_word (&msg, new_thread.raw);

	      break;
	    }
	  CASE (TASK_THREAD_DESTROY, 2)
	    {
	      struct task_object *target;
	      l4_word_t handle;
	      l4_thread_id_t victim;

	      l4_msg_store (tag, &msg);
	      handle = l4_msg_word (&msg, 1);
	      victim.raw = l4_msg_word (&msg, 2);

	      target = task_by_control_handle(caller_task, handle);
	      if (!target)
		{ err = EPERM; goto fail; }

	      err = task_thread_destroy(target, victim);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);

	      break;
	    }
	      
	  CASE (TASK_TRANSACTION_CREATE, 2)
	    {
	      /* Initiates sending HANDLE from CALLER_TASK to
		 RECEIVER_TASK. SUBJECT is the task that HANDLE refers
		 to. */
	      l4_word_t receiver;
	      l4_word_t id;
	      
	      const struct task_handle *handle;
	      const struct task_object *receiver_task;
	      l4_word_t new_transaction;

	      l4_msg_store (tag, &msg);
	      receiver = l4_msg_word (&msg, 1);
	      id = l4_msg_word (&msg, 2);

	      receiver_task = task_by_handle(caller_task, receiver);
	      if (!receiver_task)
		{ err = ESRCH; goto fail; }

	      handle = task_handle_lookup (caller_task, id);
	      if (!handle)
		{ err = ESRCH; goto fail; }

	      err = transaction_create(caller_task,
				       receiver_task,
				       handle->task,
				       handle->type,
				       &new_transaction);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      l4_msg_append_word (&msg, new_transaction);

	      break;	      
	    }
	  CASE (TASK_TRANSACTION_RECEIVE, 2)
	    {
	      /* The CALLER tries to get a handle from SENDER, using
	         the transaction id previously created by the sender */
	      l4_word_t sender;
	      l4_word_t transaction;

	      const struct task_object *sender_task;
	      l4_word_t handle;

	      l4_msg_store (tag, &msg);
	      sender = l4_msg_word (&msg, 1);
	      transaction = l4_msg_word (&msg, 2);

	      sender_task = task_by_handle(caller_task, sender);
	      if (!sender_task)
		{ err = ESRCH; goto fail; }
	      
	      err = transaction_receive(transaction,
					sender_task,
					caller_task,
					&handle);

	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	      l4_msg_append_word (&msg, handle);

	      break;
	    }
	  CASE (TASK_TRANSACTION_DESTROY, 1)
	    {
	      l4_word_t transaction;
	      
	      l4_msg_store (tag, &msg);
	      transaction = l4_msg_word (&msg, 1);

	      err = transaction_destroy(caller_task, transaction);
	      if (err)
		goto fail;

	      l4_msg_clear (&msg);
	    }
	default:
	  err = EINVAL;
	fail:
	  l4_msg_clear (&msg);
	  l4_set_msg_label (&msg, err);
	  
	  break;
	}
      
      /* Ignores return value */
      l4_msg_load (&msg);
      (void) l4_reply(caller);
    }
}

int
main (int argc, char *argv[])
{
  output_debug = 1;

  debug ("%s " PACKAGE_VERSION "\n", program_name);

  task_init();
  transaction_init();
  
  /* Only one thread, for now */
  serve();

  return 1;
}
