/* task.h - Generic definitions.
   Copyright (C) 2003 Free Software Foundation, Inc.
   Written by Marcus Brinkmann.

   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA. */

#ifndef TASK_SERVER_TASK_H
#define TASK_SERVER_TASK_H


#include <l4.h>
#include "string.h"
#include "output.h"


/* The program name.  */
extern char program_name[];

#define BUG_ADDRESS	"<bug-hurd@gnu.org>"

int main (int argc, char *argv[]);

/* FIXME: This should be the task server's cap ID.  */
#define TASK_WORTEL_CAP 1

enum task_operation
{
  /* Task handling */
  TASK_CREATE = 1,
  TASK_DESTROY,
  TASK_REVOKE,
  TASK_RELEASE,
  TASK_QUERY,
  TASK_REFER,

  /* Thread handling */
  TASK_THREAD_CREATE,
  TASK_THREAD_DESTROY,

  /* Manager */
  TASK_REGISTER_MANAGER,
  TASK_SET_ACCOUNTING,
  TASK_GROUP_DESTROY,
  TASK_CONTROL,

  /* Handle transfer */
  TASK_TRANSACTION_CREATE,
  TASK_TRANSACTION_RECEIVE,
  TASK_TRANSACTION_DESTROY,

  /* Notifications. Not really in the same name space. */
  TASK_DEATH_NOTIFICATION
};

#define TASK_CONTROL_MASK (1UL << 14)
#define TASK_MANAGER_MASK (TASK_CONTROL_MASK << 1)

/* FIXME: Should be defined in some header file */
#if L4_WORDSIZE_32
# define L4_BITS_PER_WORD 32
# define SUBSYSTEM_SHIFT 15
#elif L4_WORDSIZE_64
# define L4_BITS_PER_WORD 64
# define SUBSYSTEM_SHIFT 29
#else
# error Unknown word size
#endif

/* Place these at some random high address */
#define KIP_LOCATION (0xC000000UL)
#define UTC_LOCATION (0xD000000UL)

enum task_state {
  TASK_FREE = 0,
  TASK_EMPTY,
  TASK_ALIVE,
  TASK_ZOMBIE
};

struct task_object;

/* task-object.c prototypes */
void
task_init(void);

void
task_reference_add(l4_word_t a, l4_word_t b);

void
task_control_add(l4_word_t a, l4_word_t b);

void
task_incref(struct task_object *task);

void
task_decref(struct task_object *task);

l4_word_t
taskid_by_thread(l4_thread_id_t thread);

struct task_object *
task_by_thread(l4_thread_id_t thread);

const struct task_object *
task_by_handle(struct task_object *owner, l4_word_t id);

struct task_object *
task_by_control_handle(struct task_object *owner, l4_word_t id);

int
task_manager_p(struct task_object *owner, l4_word_t id);

l4_word_t
task_create(struct task_object *creator, l4_word_t thread_max,
	    l4_word_t *control_id, l4_word_t *ref_id);

l4_word_t
task_destroy(struct task_object *task);

l4_word_t
task_release(struct task_object *task, l4_word_t handle);

l4_word_t
task_query (const struct task_object *task,
	    l4_word_t *taskid, l4_word_t *state);

l4_word_t
task_group_destroy(l4_word_t accounting_id);

l4_word_t
task_thread_create(struct task_object *task, l4_thread_id_t pager,
		   l4_thread_id_t *new_thread);

l4_word_t
task_thread_destroy(struct task_object *task, l4_thread_id_t victim);

/* Handles/capabilities */
enum task_handle_type
{
  HANDLE_FREE, HANDLE_REF, HANDLE_CONTROL, HANDLE_MANAGER
};

struct task_handle
{
  enum task_handle_type type;
  unsigned refs;
  struct task_object *task;
};

const struct task_handle *
task_handle_lookup(struct task_object *owner, l4_word_t id);

l4_word_t
task_add_cap (struct task_object *task,
	      enum task_handle_type type,
	      struct task_object *subject,
	      l4_word_t *id);


/* task-transaction prototypes */
void
transaction_init(void);

void
transaction_note_death(const struct task_object *task);

/* Operations */
l4_word_t
transaction_create(const struct task_object *sender,
		   const struct task_object *receiver,
		   struct task_object *subject,
		   enum task_handle_type type,
		   l4_word_t *new_transaction);

l4_word_t
transaction_receive(l4_word_t id,
		    const struct task_object *sender,
		    struct task_object *receiver,
		    l4_word_t *handle);

l4_word_t
transaction_destroy(struct task_object *sender, l4_word_t id);


/* task-manager functions */

l4_word_t
task_register_manager (struct task_object *task, l4_word_t *handle);

#endif /* TASK_SERVER_TASK_H */
