
#include <errno.h>

#include "task.h"

/* Should be declared in <assert.h> */
#define assert(x)

#define TRANSACTION_MAX 1000

/* A transaction record, for sending a task handle */
struct transaction_record
{
  enum transaction_state
    { TRANSACTION_FREE = 0,
      TRANSACTION_PENDING,
      TRANSACTION_ERROR,
      TRANSACTION_DONE } state;
  enum task_handle_type type;
  const struct task_object *sender;
  const struct task_object *receiver;
  /* Subject is NULL for manager handles */
  struct task_object *subject;
};

/* A transaction record imples an automatic reference to the subject
   of the transaction. The reason is that otherwise, if the sender
   drops it's references to a zombie subject, the id can get reused,
   and the receiver ends up with a handle to a random new task, which
   the sender never had any right to give away.

   The current code reliably delivers zombie handles to the receiver,
   both for zombies the sender pass in, and for tasks that turn into
   zombies while the handle is in transit.

   An alternative is to forbid transfer of zombie handles, invalidate
   the transaction record whenever the subject turns into a zombie.

   A transaction record does not imply any reference to the sender or
   receiver; if the receiver dies, the record state changes to
   TRANSACTION_ERROR. If the sender dies, the record is deallocated
   immediately. */

/* A transaction id is an index into this array. Index zero is not
   used. */
static struct transaction_record
transaction[TRANSACTION_MAX];

void
transaction_init(void)
{
  /* Redundant if linker/loader does its job */
  memset(transaction, 0, sizeof(transaction));
}

static l4_word_t
transaction_alloc(void)
{
  /* Inefficient linear search */
  l4_word_t id;
  for (id = 1; id < TRANSACTION_MAX; id++)
    if (transaction[id].state == TRANSACTION_FREE)
      return id;

  return 0;
}


/* Called when task id turns into a zombie */
void
transaction_note_death(const struct task_object *task)
{
  l4_word_t id;
  for (id = 1; id < TRANSACTION_MAX; id++)
    if (transaction[id].state != TRANSACTION_FREE)
      {
	/* If the sender is gone, the transaction is deallocated
	   immediately */
	if (transaction[id].sender == task)
	  {
	    if (transaction[id].state == TRANSACTION_PENDING
		&& transaction[id].subject)
	      task_decref(transaction[id].subject);
	  
	    transaction[id].state = TRANSACTION_FREE;
	    continue;
	  }
	/* Receiver death matters only for pending transactions */
	else if (transaction[id].state == TRANSACTION_PENDING
		 && transaction[id].receiver == task)
	  {
	    transaction[id].state = TRANSACTION_ERROR;
	    if (transaction[id].subject)
	      task_decref(transaction[id].subject);
	  }
      }
}

/* Operations */

l4_word_t
transaction_create(const struct task_object *sender,
		   const struct task_object *receiver,
		   struct task_object *subject,
		   enum task_handle_type type,
		   l4_word_t *new_transaction)
{
  l4_word_t id = transaction_alloc();
  if (!id)
    return ENOMEM;

  if (subject)
    task_incref(subject);

  transaction[id].state = TRANSACTION_PENDING;
  transaction[id].type = type;
  transaction[id].sender = sender;
  transaction[id].receiver = receiver;
  transaction[id].subject = subject;
  
  *new_transaction = id;
  return 0;
}

l4_word_t
transaction_receive(l4_word_t id,
		    const struct task_object *sender,
		    struct task_object *receiver,
		    l4_word_t *handle)
{
  if (id >= TRANSACTION_MAX
      || transaction[id].state == TRANSACTION_FREE)
    return EINVAL;

  /* It's not useful to distinguish between these errors,
     as any state is possible if the sender has destroyed the record. */
  if (id && id < TRANSACTION_MAX
      && transaction[id].state == TRANSACTION_PENDING
      && transaction[id].sender == sender
      && transaction[id].receiver == receiver)
    {
      l4_word_t err = task_add_cap (receiver,
				    transaction[id].type,
				    transaction[id].subject, handle);

      transaction[id].state = err ? TRANSACTION_ERROR : TRANSACTION_DONE;
      return err;
    }
  else
    return EINVAL;
}

l4_word_t
transaction_destroy(struct task_object *sender, l4_word_t id)
{
  enum transaction_state state;
  
  if (id >= TRANSACTION_MAX
      || transaction[id].state == TRANSACTION_FREE
      || transaction[id].sender != sender)
    return EINVAL;

  state = transaction[id].state;
  transaction[id].state = TRANSACTION_FREE;

  switch (state)
    {
    case TRANSACTION_PENDING:
      if (transaction[id].subject)
	task_decref(transaction[id].subject);
      return EAGAIN;
    case TRANSACTION_ERROR:
      return EPIPE;
    case TRANSACTION_DONE:
      return 0;
    default:
      panic ("transaction_destroy: Invalid state.");
      return EINVAL;
    }
}
